@@ -1,6 +1,6 @@
 #! /bin/sh
 #
-# This is a kernel build script for openSUSE 10.3's 2.6.22 kernel.
+# This is a kernel build script for openSUSE 10.2's 2.6.18 kernel.
 #
 
 die () {
@@ -60,53 +60,46 @@ fi
 
 cd /tmp/ || die "Can't chdir to /tmp/ ."
 
-if [ ! -r kernel-source-2.6.22.18-0.2.src.rpm ]
+if [ ! -r kernel-source-2.6.18.8-0.13.src.rpm ]
 then
-    wget http://download.opensuse.org/update/10.3/rpm/src/kernel-source-2.6.22.18-0.2.src.rpm || die "Can't download source package."
+    wget http://ftp.riken.jp/Linux/suse/suse/update/10.2/rpm/src/kernel-source-2.6.18.8-0.13.src.rpm || die "Can't download source package."
 fi
-rpm -ivh kernel-source-2.6.22.18-0.2.src.rpm || die "Can't install source package."
+rpm -ivh kernel-source-2.6.18.8-0.13.src.rpm || die "Can't install source package."
 
 cd /usr/src/packages/SOURCES/ || die "Can't chdir to /usr/src/packages/SOURCES/ ."
-if [ ! -r ccs-patch-1.6.4-20080903.tar.gz ]
+if [ ! -r ccs-patch-1.6.5-20081210.tar.gz ]
 then
-    wget http://osdn.dl.sourceforge.jp/tomoyo/30297/ccs-patch-1.6.4-20080903.tar.gz || die "Can't download patch."
+    wget http://osdn.dl.sourceforge.jp/tomoyo/30297/ccs-patch-1.6.5-20081210.tar.gz || die "Can't download patch."
 fi
 
 cd /tmp/ || die "Can't chdir to /tmp/ ."
 cp -p /usr/src/packages/SOURCES/kernel-default.spec . || die "Can't copy spec file."
 patch << "EOF" || die "Can't patch spec file."
---- kernel-default.spec	2008-06-10 00:07:47.000000000 +0900
-+++ kernel-default.spec	2008-06-20 09:39:20.000000000 +0900
-@@ -37,10 +37,10 @@
- %define build_vanilla 1
- %endif
+--- kernel-default.spec	2008-10-17 00:22:46.000000000 +0900
++++ kernel-default.spec	2008-10-18 14:36:07.000000000 +0900
+@@ -30,13 +30,13 @@
+ %define build_um  %([ default != um ] ; echo $?)
+ %define build_vanilla  %([ default != vanilla ] ; echo $?)
  
 -Name:           kernel-default
 +Name:           ccs-kernel-default
+ %ifarch ia64
+ # arch/ia64/scripts/unwcheck.py
+ BuildRequires:  python
+ %endif
+ Version:        2.6.18.8
+-Release: 0.13
++Release: 0.13_tomoyo_1.6.5
  Summary:        The Standard Kernel for both Uniprocessor and Multiprocessor Systems
- Version:        2.6.22.18
--Release: 0.2
-+Release: 0.2_tomoyo_1.6.4
  License:        GPL v2 or later
  Group:          System/Kernel
- AutoReqProv:    on
-@@ -191,7 +191,7 @@
- %define tolerate_unknown_new_config_options 0
- # kABI change tolerance (default in maintenance should be 4, 6, 8 or 15,
- # 31 is the maximum; see scripts/kabi-checks)
--%define tolerate_kabi_changes 6
-+%define tolerate_kabi_changes 31
- 
- %description
- The standard kernel for both uniprocessor and multiprocessor systems.
-@@ -281,6 +281,11 @@
+@@ -242,6 +242,10 @@
  %build
  source .rpm-defs
- cd linux-2.6.22
+ cd linux-2.6.18
 +# TOMOYO Linux
-+tar -zxf %_sourcedir/ccs-patch-1.6.4-20080903.tar.gz
-+patch -sp1 < patches/ccs-patch-2.6.22-suse-10.3.diff
-+# sed -i -e 's:-ccs::' -- Makefile
++tar -zxf %_sourcedir/ccs-patch-1.6.5-20081210.tar.gz
++patch -sp1 < patches/ccs-patch-2.6.18-suse-10.2.diff
 +cat config.ccs >> .config
  cp .config .config.orig
  %if %{tolerate_unknown_new_config_options}
