@@ -5,7 +5,7 @@
  *
  * Copyright (C) 2005-2008  NTT DATA CORPORATION
  *
- * Version: 1.6.4   2008/09/03
+ * Version: 1.6.5   2008/11/11
  *
  * This file is applicable to both 2.4.30 and 2.6.11 and later.
  * See README.ccs for ChangeLog.
@@ -40,6 +40,7 @@
 struct dentry;
 struct vfsmount;
 struct in6_addr;
+extern asmlinkage long sys_getpid(void);
 extern asmlinkage long sys_getppid(void);
 
 #define false 0
@@ -88,6 +89,48 @@ extern asmlinkage long sys_getppid(void);
 			ret; })
 #endif
 
+#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 0)
+#define smp_read_barrier_depends smp_rmb
+#endif
+
+#ifndef ACCESS_ONCE
+#define ACCESS_ONCE(x) (*(volatile typeof(x) *)&(x))
+#endif
+
+#ifndef rcu_dereference
+#define rcu_dereference(p)     ({ \
+				typeof(p) _________p1 = ACCESS_ONCE(p); \
+				smp_read_barrier_depends(); /* see RCU */ \
+				(_________p1); \
+				})
+#endif
+
+#ifndef rcu_assign_pointer
+#define rcu_assign_pointer(p, v) \
+	({ \
+		if (!__builtin_constant_p(v) || \
+		    ((v) != NULL)) \
+			smp_wmb(); /* see RCU */ \
+		(p) = (v); \
+	})
+#endif
+
+#ifndef list_for_each_rcu
+#define list_for_each_rcu(pos, head) \
+	for (pos = rcu_dereference((head)->next); \
+		prefetch(pos->next), pos != (head); \
+		pos = rcu_dereference(pos->next))
+#endif
+
+#ifndef list_for_each_entry_rcu
+#define list_for_each_entry_rcu(pos, head, member) \
+	for (pos = list_entry(rcu_dereference((head)->next), typeof(*pos), \
+		member); \
+		prefetch(pos->member.next), &pos->member != (head); \
+		pos = list_entry(rcu_dereference(pos->member.next), \
+		typeof(*pos), member))
+#endif
+
 /*
  * Singly linked list.
  *
@@ -111,33 +154,14 @@ static inline void INIT_LIST1_HEAD(struct list1_head *list)
 	list->next = list;
 }
 
-/**
- * list1_entry - get the struct for this entry
- * @ptr:        the &struct list1_head pointer.
- * @type:       the type of the struct this is embedded in.
- * @member:     the name of the list1_struct within the struct.
- */
-#define list1_entry(ptr, type, member) container_of(ptr, type, member)
+/* Reuse list_entry because it doesn't use "->prev" pointer. */
+#define list1_entry list_entry
 
-/**
- * list1_for_each        -       iterate over a list
- * @pos:        the &struct list1_head to use as a loop cursor.
- * @head:       the head for your list.
- */
-#define list1_for_each(pos, head)					\
-	for (pos = (head)->next; prefetch(pos->next), pos != (head);	\
-	     pos = pos->next)
+/* Reuse list_for_each_rcu because it doesn't use "->prev" pointer. */
+#define list1_for_each list_for_each_rcu
 
-/**
- * list1_for_each_entry  -       iterate over list of given type
- * @pos:        the type * to use as a loop cursor.
- * @head:       the head for your list.
- * @member:     the name of the list1_struct within the struct.
- */
-#define list1_for_each_entry(pos, head, member)				\
-	for (pos = list1_entry((head)->next, typeof(*pos), member);	\
-	     prefetch(pos->member.next), &pos->member != (head);        \
-	     pos = list1_entry(pos->member.next, typeof(*pos), member))
+/* Reuse list_for_each_entry_rcu because it doesn't use "->prev" pointer. */
+#define list1_for_each_entry list_for_each_entry_rcu
 
 /**
  * list1_for_each_cookie - iterate over a list with cookie.
@@ -145,14 +169,20 @@ static inline void INIT_LIST1_HEAD(struct list1_head *list)
  * @cookie:     the &struct list1_head to use as a cookie.
  * @head:       the head for your list.
  *
- * Same with list_for_each except that this primitive uses cookie
+ * Same with list_for_each_rcu() except that this primitive uses @cookie
  * so that we can continue iteration.
+ * @cookie must be NULL when iteration starts, and @cookie will become
+ * NULL when iteration finishes.
+ *
+ * Since list elements are never removed, we don't need to get a lock
+ * or a reference count.
  */
-#define list1_for_each_cookie(pos, cookie, head)			\
-	for (({ if (!cookie)						\
-				     cookie = head; }), pos = (cookie)->next; \
-	     prefetch(pos->next), pos != (head) || ((cookie) = NULL);	\
-	     (cookie) = pos, pos = pos->next)
+#define list1_for_each_cookie(pos, cookie, head)                      \
+	for (({ if (!cookie)                                          \
+				     cookie = head; }),               \
+	     pos = rcu_dereference((cookie)->next);                   \
+	     prefetch(pos->next), pos != (head) || ((cookie) = NULL); \
+	     (cookie) = pos, pos = rcu_dereference(pos->next))
 
 /**
  * list_add_tail_mb - add a new entry with memory barrier.
@@ -160,18 +190,17 @@ static inline void INIT_LIST1_HEAD(struct list1_head *list)
  * @head: list head to add it before.
  *
  * Same with list_add_tail_rcu() except that this primitive uses mb()
- * so that we can traverse forwards using list_for_each() and
- * list_for_each_cookie().
+ * so that we can traverse forwards using list1_for_each() and
+ * list1_for_each_cookie().
  */
 static inline void list1_add_tail_mb(struct list1_head *new,
 				     struct list1_head *head)
 {
-	struct list1_head *pos = head;
+	struct list1_head *prev = head;
 	new->next = head;
-	mb(); /* Avoid out-of-order execution. */
-	while (pos->next != head)
-		pos = pos->next;
-	pos->next = new;
+	while (prev->next != head)
+		prev = prev->next;
+	rcu_assign_pointer(prev->next, new);
 }
 
 /* Temporary buffer for holding pathnames. */
@@ -200,10 +229,32 @@ struct obj_info {
 	/* I don't handle path2_stat for rename operation. */
 	struct mini_stat path1_parent_stat;
 	struct mini_stat path2_parent_stat;
-	struct linux_binprm *bprm;
 	struct ccs_page_buffer *tmp;
 };
 
+/* Structure for " if " and "; set" part. */
+struct condition_list {
+	struct list1_head list;
+	u16 condc;
+	u16 argc;
+	u16 envc;
+	u8 post_state[4];
+	/* "unsigned long condition[condc]" follows here. */
+	/* "struct argv_entry argv[argc]" follows here. */
+	/* "struct envp_entry envp[envc]" follows here. */
+};
+
+/* Structure for request info. */
+struct ccs_request_info {
+	struct domain_info *domain;
+	struct linux_binprm *bprm;
+	struct obj_info *obj;
+	const struct condition_list *cond;
+	u16 retry;
+	u8 profile;
+	u8 mode;
+};
+
 /* Structure for holding a token. */
 struct path_info {
 	const char *name;
@@ -309,7 +360,7 @@ struct domain_info {
 /* Ignore "allow_env" directive in exception policy.  */
 #define DOMAIN_FLAGS_IGNORE_GLOBAL_ALLOW_ENV  2
 /*
- * This domain was unable to create a new domain at ccs_find_next_domain()
+ * This domain was unable to create a new domain at find_next_domain()
  * because the name of the domain to be created was too long or
  * it could not allocate memory.
  * More than one process continued execve() without domain transition.
@@ -527,10 +578,8 @@ struct ccs_io_buffer {
 	int (*read) (struct ccs_io_buffer *);
 	int (*write) (struct ccs_io_buffer *);
 	int (*poll) (struct file *file, poll_table *wait);
-	/* Exclusive lock for read_buf.         */
-	struct mutex read_sem;
-	/* Exclusive lock for write_buf.        */
-	struct mutex write_sem;
+	/* Exclusive lock for this structure.   */
+	struct mutex io_sem;
 	/* The position currently reading from. */
 	struct list1_head *read_var1;
 	/* Extra variables for reading.         */
@@ -543,6 +592,8 @@ struct ccs_io_buffer {
 	char *read_buf;
 	/* EOF flag for reading.                */
 	bool read_eof;
+	/* Read domain ACL of specified PID?    */
+	bool read_single_domain;
 	/* Extra variable for reading.          */
 	u8 read_bit;
 	/* Bytes available for reading.         */
@@ -561,8 +612,8 @@ struct ccs_io_buffer {
 struct condition_list;
 
 /* Check conditional part of an ACL entry. */
-bool ccs_check_condition(const struct acl_info *acl,
-			 struct obj_info *obj_info);
+bool ccs_check_condition(struct ccs_request_info *r,
+			 const struct acl_info *acl);
 /* Check whether the domain has too many ACL entries to hold. */
 bool ccs_check_domain_quota(struct domain_info * const domain);
 /* Transactional sprintf() for policy dump. */
@@ -617,10 +668,9 @@ bool ccs_read_pivot_root_policy(struct ccs_io_buffer *head);
 /* Read "deny_autobind" entry in system policy. */
 bool ccs_read_reserved_port_policy(struct ccs_io_buffer *head);
 /* Write domain policy violation warning message to console? */
-bool ccs_verbose_mode(void);
+bool ccs_verbose_mode(const struct domain_info *domain);
 /* Allocate buffer for domain policy auditing. */
-char *ccs_init_audit_log(int *len, const u8 profile, const u8 mode,
-			 struct linux_binprm *bprm);
+char *ccs_init_audit_log(int *len, struct ccs_request_info *r);
 /* Convert capability index to capability name. */
 const char *ccs_cap2keyword(const u8 operation);
 /* Convert double path operation to operation name. */
@@ -635,18 +685,16 @@ const char *ccs_get_msg(const bool is_enforce);
 const char *ccs_net2keyword(const u8 operation);
 /* Convert single path operation to operation name. */
 const char *ccs_sp2keyword(const u8 operation);
+/* Fetch next_domain from the list. */
+struct domain_info *ccs_fetch_next_domain(void);
 /* Create conditional part of an ACL entry. */
 const struct condition_list *
 ccs_find_or_assign_new_condition(char * const condition);
-/* Read conditional part of an ACL entry. */
-const struct condition_list *
-ccs_get_condition_part(const struct acl_info *acl);
 /* Add an ACL entry to domain's ACL list. */
 int ccs_add_domain_acl(struct domain_info *domain, struct acl_info *acl);
-/* Check whether there is space for audit logs. */
-int ccs_can_save_audit_log(const bool is_granted);
 /* Ask supervisor's opinion. */
-int ccs_check_supervisor(struct linux_binprm *bprm, const char *fmt, ...)
+int ccs_check_supervisor(struct ccs_request_info *r,
+			 const char *fmt, ...)
 	__attribute__ ((format(printf, 2, 3)));
 /* Close /proc/ccs/ interface. */
 int ccs_close_control(struct file *file);
@@ -680,7 +728,9 @@ int ccs_write_argv0_policy(char *data, struct domain_info *domain,
 			   const struct condition_list *condition,
 			   const bool is_delete);
 /* Write an audit log. */
-int ccs_write_audit_log(char *log, const bool is_granted);
+int ccs_write_audit_log(const bool is_granted, struct ccs_request_info *r,
+			const char *fmt, ...)
+     __attribute__ ((format(printf, 3, 4)));
 /* Create "allow_capability" entry in domain policy. */
 int ccs_write_capability_policy(char *data, struct domain_info *domain,
 				const struct condition_list *condition,
@@ -747,17 +797,18 @@ struct domain_info *ccs_find_or_assign_new_domain(const char *domainname,
 						  const u8 profile);
 /* Undelete a domain. */
 struct domain_info *ccs_undelete_domain(const char *domainname);
-/* Check mode for specified capability. */
-u8 ccs_check_capability_flags(const u8 index);
 /* Check mode for specified functionality. */
-unsigned int ccs_check_flags(const u8 index);
-/* Same with ccs_check_flags() except that it doesn't check might_sleep(). */
-unsigned int ccs_check_flags_no_sleep_check(const u8 index);
+unsigned int ccs_check_flags(const struct domain_info *domain, const u8 index);
+/* Check whether it is safe to sleep. */
+bool ccs_can_sleep(void);
 /* Allocate memory for structures. */
 void *ccs_alloc_acl_element(const u8 acl_type,
 			    const struct condition_list *condition);
 /* Fill in "struct path_info" members. */
 void ccs_fill_path_info(struct path_info *ptr);
+/* Fill in "struct ccs_request_info" members. */
+void ccs_init_request_info(struct ccs_request_info *r,
+			   struct domain_info *domain, const u8 index);
 /* Run policy loader when /sbin/init starts. */
 void ccs_load_policy(const char *filename);
 /* Print an IPv6 address. */
@@ -766,11 +817,18 @@ void ccs_print_ipv6(char *buffer, const int buffer_len,
 /* Change "struct domain_info"->flags. */
 void ccs_set_domain_flag(struct domain_info *domain, const bool is_delete,
 			 const u8 flags);
-/* Update the process's state. */
-void ccs_update_condition(const struct acl_info *acl);
 /* Update the policy change counter. */
 void ccs_update_counter(const unsigned char index);
 
+/* Check whether the basename of program and argv0 is allowed to differ. */
+int ccs_check_argv0_perm(struct ccs_request_info *r,
+			 const struct path_info *filename, const char *argv0);
+/* Check whether the given environment is allowed to be received. */
+int ccs_check_env_perm(struct ccs_request_info *r, const char *env);
+/* Check whether the given pathname is allowed to be executed. */
+int ccs_check_exec_perm(struct ccs_request_info *r,
+			const struct path_info *filename);
+
 /* strcmp() for "struct path_info" structure. */
 static inline bool ccs_pathcmp(const struct path_info *a,
 			       const struct path_info *b)
@@ -790,6 +848,20 @@ static inline u8 ccs_acl_type2(struct acl_info *ptr)
 	return ptr->type & ~ACL_WITH_CONDITION;
 }
 
+/**
+ * ccs_get_condition_part - Get condition part of the given ACL entry.
+ *
+ * @acl: Pointer to "struct acl_info". Pointer to an ACL entry.
+ *
+ * Returns pointer to the condition part if the ACL has it, NULL otherwise.
+ */
+static inline const struct condition_list *
+ccs_get_condition_part(const struct acl_info *acl)
+{
+	return (acl->type & ACL_WITH_CONDITION) ?
+		acl->access_me_via_ccs_get_condition_part : NULL;
+}
+
 /* A linked list of domains. */
 extern struct list1_head domain_list;
 /* Has /sbin/init started? */
@@ -798,8 +870,6 @@ extern bool sbin_init_started;
 extern const char *ccs_log_level;
 /* The kernel's domain. */
 extern struct domain_info KERNEL_DOMAIN;
-/* Exclusive lock for updating domain policy. */
-extern struct mutex domain_acl_lock;
 
 #include <linux/dcache.h>
 extern spinlock_t vfsmount_lock;
