@@ -5,7 +5,7 @@
  *
  * Copyright (C) 2005-2008  NTT DATA CORPORATION
  *
- * Version: 1.6.1   2008/05/10
+ * Version: 1.6.5   2008/11/11
  *
  * This file is applicable to both 2.4.30 and 2.6.11 and later.
  * See README.ccs for ChangeLog.
@@ -303,6 +303,7 @@ static inline int ccs_socket_sendmsg_permission(struct socket *sock,
 }
 
 #if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 22)
+#if !defined(RHEL_MAJOR) || RHEL_MAJOR != 5
 
 static inline struct iphdr *ip_hdr(const struct sk_buff *skb)
 {
@@ -319,6 +320,7 @@ static inline struct ipv6hdr *ipv6_hdr(const struct sk_buff *skb)
 	return skb->nh.ipv6h;
 }
 
+#endif
 #endif
 
 /*
@@ -455,13 +457,4 @@ static inline int ccs_socket_recv_datagram_permission(struct sock *sk,
 
 #endif
 
-/* For compatibility with 1.4.x/1.5.x patches */
-#define CheckSocketSendMsgPermission      ccs_socket_sendmsg_permission
-#define CheckSocketCreatePermission       ccs_socket_create_permission
-#define CheckSocketBindPermission         ccs_socket_bind_permission
-#define CheckSocketListenPermission       ccs_socket_listen_permission
-#define CheckSocketAcceptPermission       ccs_socket_accept_permission
-#define CheckSocketConnectPermission      ccs_socket_connect_permission
-#define CheckSocketRecvDatagramPermission ccs_socket_recv_datagram_permission
-
 #endif
