@@ -5,7 +5,7 @@
  *
  * Copyright (C) 2005-2008  NTT DATA CORPORATION
  *
- * Version: 1.6.0   2008/04/01
+ * Version: 1.6.5   2008/11/11
  *
  * This file is applicable to both 2.4.30 and 2.6.11 and later.
  * See README.ccs for ChangeLog.
@@ -37,11 +37,8 @@ LIST1_HEAD(domain_list);
 
 #ifdef CONFIG_TOMOYO
 
-/* Lock for appending domain's ACL. */
-DEFINE_MUTEX(domain_acl_lock);
-
 /* Domain creation lock. */
-static DEFINE_MUTEX(new_domain_assign_lock);
+static DEFINE_MUTEX(domain_list_lock);
 
 /* Structure for "initialize_domain" and "no_initialize_domain" keyword. */
 struct domain_initializer_entry {
@@ -91,12 +88,16 @@ struct alias_entry {
 void ccs_set_domain_flag(struct domain_info *domain, const bool is_delete,
 			 const u8 flags)
 {
-	mutex_lock(&new_domain_assign_lock);
+	/* We need to serialize because this is bitfield operation. */
+	static DEFINE_SPINLOCK(lock);
+	/***** CRITICAL SECTION START *****/
+	spin_lock(&lock);
 	if (!is_delete)
 		domain->flags |= flags;
 	else
 		domain->flags &= ~flags;
-	mutex_unlock(&new_domain_assign_lock);
+	spin_unlock(&lock);
+	/***** CRITICAL SECTION END *****/
 }
 
 /**
@@ -108,7 +109,8 @@ void ccs_set_domain_flag(struct domain_info *domain, const bool is_delete,
  */
 const char *ccs_get_last_name(const struct domain_info *domain)
 {
-	const char *cp0 = domain->domainname->name, *cp1 = strrchr(cp0, ' ');
+	const char *cp0 = domain->domainname->name;
+	const char *cp1 = strrchr(cp0, ' ');
 	if (cp1)
 		return cp1 + 1;
 	return cp0;
@@ -124,10 +126,20 @@ const char *ccs_get_last_name(const struct domain_info *domain)
  */
 int ccs_add_domain_acl(struct domain_info *domain, struct acl_info *acl)
 {
-	if (domain)
+	if (domain) {
+		/*
+		 * We need to serialize because this function is called by
+		 * various update functions.
+		 */
+		static DEFINE_SPINLOCK(lock);
+		/***** CRITICAL SECTION START *****/
+		spin_lock(&lock);
 		list1_add_tail_mb(&acl->list, &domain->acl_info_list);
-	else
+		spin_unlock(&lock);
+		/***** CRITICAL SECTION END *****/
+	} else {
 		acl->type &= ~ACL_DELETED;
+	}
 	ccs_update_counter(CCS_UPDATES_COUNTER_DOMAIN_POLICY);
 	return 0;
 }
@@ -160,58 +172,26 @@ static int audit_execute_handler_log(const bool is_default,
 				     const char *handler,
 				     struct linux_binprm *bprm)
 {
-	char *buf;
-	int len;
-	int len2;
-	u8 profile;
-	u8 mode;
-	if (ccs_can_save_audit_log(true) < 0)
-		return -ENOMEM;
-	len = strlen(handler) + 32;
-	profile = current->domain_info->profile;
-	mode = ccs_check_flags(CCS_TOMOYO_MAC_FOR_FILE);
-	buf = ccs_init_audit_log(&len, profile, mode, bprm);
-	if (!buf)
-		return -ENOMEM;
-	len2 = strlen(buf);
-	snprintf(buf + len2, len - len2 - 1, "%s %s\n",
-		 is_default ? KEYWORD_EXECUTE_HANDLER :
-		 KEYWORD_DENIED_EXECUTE_HANDLER, handler);
-	return ccs_write_audit_log(buf, true);
+	struct ccs_request_info r;
+	ccs_init_request_info(&r, NULL, CCS_TOMOYO_MAC_FOR_FILE);
+	r.bprm = bprm;
+	return ccs_write_audit_log(true, &r, "%s %s\n",
+				   is_default ? KEYWORD_EXECUTE_HANDLER :
+				   KEYWORD_DENIED_EXECUTE_HANDLER, handler);
 }
 
 /**
  * audit_domain_creation_log - Audit domain creation log.
  *
- * @domainname: The name of newly created domain.
- * @mode:       Access control mode used.
- * @profile:    Profile number used.
+ * @domain:  Pointer to "struct domain_info".
  *
  * Returns 0 on success, negative value otherwise.
  */
-static int audit_domain_creation_log(const char *domainname, const u8 mode,
-				     const u8 profile)
+static int audit_domain_creation_log(struct domain_info *domain)
 {
-	char *buf;
-	char *cp;
-	int len;
-	int len2;
-	if (ccs_can_save_audit_log(false) < 0)
-		return -ENOMEM;
-	len = strlen(domainname) + 32;
-	buf = ccs_init_audit_log(&len, profile, mode, NULL);
-	if (!buf)
-		return -ENOMEM;
-	cp = strchr(buf, '\n');
-	if (!cp) {
-		ccs_free(buf);
-		return -ENOMEM;
-	}
-	*++cp = '\0';
-	len2 = strlen(buf);
-	snprintf(buf + len2, len - len2 - 1, "%s\nuse_profile %u\n",
-		 domainname, profile);
-	return ccs_write_audit_log(buf, false);
+	struct ccs_request_info r;
+	ccs_init_request_info(&r, domain, CCS_TOMOYO_MAC_FOR_FILE);
+	return ccs_write_audit_log(false, &r, "use_profile %u\n", r.profile);
 }
 
 /* The list for "struct domain_initializer_entry". */
@@ -353,7 +333,7 @@ static bool is_domain_initializer(const struct path_info *domainname,
 {
 	struct domain_initializer_entry *ptr;
 	bool flag = false;
-	list1_for_each_entry(ptr,  &domain_initializer_list, list) {
+	list1_for_each_entry(ptr, &domain_initializer_list, list) {
 		if (ptr->is_deleted)
 			continue;
 		if (ptr->domainname) {
@@ -741,7 +721,7 @@ int ccs_delete_domain(char *domainname)
 	struct path_info name;
 	name.name = domainname;
 	ccs_fill_path_info(&name);
-	mutex_lock(&new_domain_assign_lock);
+	mutex_lock(&domain_list_lock);
 #ifdef DEBUG_DOMAIN_UNDELETE
 	printk(KERN_DEBUG "ccs_delete_domain %s\n", domainname);
 	list1_for_each_entry(domain, &domain_list, list) {
@@ -779,7 +759,7 @@ int ccs_delete_domain(char *domainname)
 #endif
 		break;
 	}
-	mutex_unlock(&new_domain_assign_lock);
+	mutex_unlock(&domain_list_lock);
 	return 0;
 }
 
@@ -797,7 +777,7 @@ struct domain_info *ccs_undelete_domain(const char *domainname)
 	struct path_info name;
 	name.name = domainname;
 	ccs_fill_path_info(&name);
-	mutex_lock(&new_domain_assign_lock);
+	mutex_lock(&domain_list_lock);
 #ifdef DEBUG_DOMAIN_UNDELETE
 	printk(KERN_DEBUG "ccs_undelete_domain %s\n", domainname);
 	list1_for_each_entry(domain, &domain_list, list) {
@@ -828,7 +808,7 @@ struct domain_info *ccs_undelete_domain(const char *domainname)
 		printk(KERN_DEBUG "%p was undeleted.\n", candidate_domain);
 #endif
 	}
-	mutex_unlock(&new_domain_assign_lock);
+	mutex_unlock(&domain_list_lock);
 	return candidate_domain;
 }
 
@@ -845,7 +825,7 @@ struct domain_info *ccs_find_or_assign_new_domain(const char *domainname,
 {
 	struct domain_info *domain = NULL;
 	const struct path_info *saved_domainname;
-	mutex_lock(&new_domain_assign_lock);
+	mutex_lock(&domain_list_lock);
 	domain = ccs_find_domain(domainname);
 	if (domain)
 		goto out;
@@ -882,11 +862,7 @@ struct domain_info *ccs_find_or_assign_new_domain(const char *domainname,
 		list1_for_each_entry(ptr, &domain->acl_info_list, list) {
 			ptr->type |= ACL_DELETED;
 		}
-		/*
-		 * Don't use ccs_set_domain_flag() because
-		 * new_domain_assign_lock is held.
-		 */
-		domain->flags = 0;
+		ccs_set_domain_flag(domain, true, domain->flags);
 		domain->profile = profile;
 		domain->quota_warned = false;
 		mb(); /* Avoid out-of-order execution. */
@@ -902,7 +878,7 @@ struct domain_info *ccs_find_or_assign_new_domain(const char *domainname,
 		list1_add_tail_mb(&domain->list, &domain_list);
 	}
  out:
-	mutex_unlock(&new_domain_assign_lock);
+	mutex_unlock(&domain_list_lock);
 	return domain;
 }
 
@@ -910,7 +886,7 @@ struct domain_info *ccs_find_or_assign_new_domain(const char *domainname,
  * get_argv0 - Get argv[0].
  *
  * @bprm: Pointer to "struct linux_binprm".
- * @tmp:  Buffer for temporal use.
+ * @tmp:  Buffer for temporary use.
  *
  * Returns true on success, false otherwise.
  */
@@ -985,35 +961,33 @@ static bool get_argv0(struct linux_binprm *bprm, struct ccs_page_buffer *tmp)
 /**
  * find_next_domain - Find a domain.
  *
- * @bprm:           Pointer to "struct linux_binprm".
- * @next_domain:    Pointer to pointer to "struct domain_info".
- * @path_to_verify: Pathname to verify. May be NULL.
- * @tmp:            Buffer for temporal use.
+ * @r:       Pointer to "struct ccs_request_info".
+ * @handler: Pathname to verify. May be NULL.
  *
  * Returns 0 on success, negative value otherwise.
  */
-static int find_next_domain(struct linux_binprm *bprm,
-			    struct domain_info **next_domain,
-			    const struct path_info *path_to_verify,
-			    struct ccs_page_buffer *tmp)
+static int find_next_domain(struct ccs_request_info *r,
+			    const struct path_info *handler)
 {
 	/*
 	 * This function assumes that the size of buffer returned by
 	 * ccs_realpath() = CCS_MAX_PATHNAME_LEN.
 	 */
-	struct domain_info *old_domain = current->domain_info;
 	struct domain_info *domain = NULL;
-	const char *old_domain_name = old_domain->domainname->name;
+	const char *old_domain_name = r->domain->domainname->name;
+	struct linux_binprm *bprm = r->bprm;
+	struct ccs_page_buffer *tmp = r->obj->tmp;
 	const char *original_name = bprm->filename;
+	const u8 mode = r->mode;
+	const bool is_enforce = (mode == 3);
+	const u32 tomoyo_flags = current->tomoyo_flags;
 	char *new_domain_name = NULL;
 	char *real_program_name = NULL;
 	char *symlink_program_name = NULL;
-	const u8 mode = ccs_check_flags(CCS_TOMOYO_MAC_FOR_FILE);
-	const bool is_enforce = (mode == 3);
+	struct path_info rn; /* real name */
+	struct path_info sn; /* symlink name */
+	struct path_info ln; /* last name */
 	int retval;
-	struct path_info r; /* real name */
-	struct path_info s; /* symlink name */
-	struct path_info l; /* last name */
 
 	{
 		/*
@@ -1030,31 +1004,36 @@ static int find_next_domain(struct linux_binprm *bprm,
 		}
 	}
 
+ retry:
+	current->tomoyo_flags = tomoyo_flags;
+	r->cond = NULL;
 	/* Get ccs_realpath of program. */
 	retval = -ENOENT; /* I hope ccs_realpath() won't fail with -ENOMEM. */
+	ccs_free(real_program_name);
 	real_program_name = ccs_realpath(original_name);
 	if (!real_program_name)
 		goto out;
 	/* Get ccs_realpath of symbolic link. */
+	ccs_free(symlink_program_name);
 	symlink_program_name = ccs_realpath_nofollow(original_name);
 	if (!symlink_program_name)
 		goto out;
 
-	r.name = real_program_name;
-	ccs_fill_path_info(&r);
-	s.name = symlink_program_name;
-	ccs_fill_path_info(&s);
-	l.name = ccs_get_last_name(old_domain);
-	ccs_fill_path_info(&l);
+	rn.name = real_program_name;
+	ccs_fill_path_info(&rn);
+	sn.name = symlink_program_name;
+	ccs_fill_path_info(&sn);
+	ln.name = ccs_get_last_name(r->domain);
+	ccs_fill_path_info(&ln);
 
-	if (path_to_verify) {
-		if (ccs_pathcmp(&r, path_to_verify)) {
+	if (handler) {
+		if (ccs_pathcmp(&rn, handler)) {
 			/* Failed to verify execute handler. */
 			static u8 counter = 20;
 			if (counter) {
 				counter--;
 				printk(KERN_WARNING "Failed to verify: %s\n",
-				       path_to_verify->name);
+				       handler->name);
 			}
 			goto out;
 		}
@@ -1062,24 +1041,25 @@ static int find_next_domain(struct linux_binprm *bprm,
 	}
 
 	/* Check 'alias' directive. */
-	if (ccs_pathcmp(&r, &s)) {
+	if (ccs_pathcmp(&rn, &sn)) {
 		struct alias_entry *ptr;
 		/* Is this program allowed to be called via symbolic links? */
 		list1_for_each_entry(ptr, &alias_list, list) {
 			if (ptr->is_deleted ||
-			    ccs_pathcmp(&r, ptr->original_name) ||
-			    ccs_pathcmp(&s, ptr->aliased_name))
+			    ccs_pathcmp(&rn, ptr->original_name) ||
+			    ccs_pathcmp(&sn, ptr->aliased_name))
 				continue;
 			memset(real_program_name, 0, CCS_MAX_PATHNAME_LEN);
 			strncpy(real_program_name, ptr->aliased_name->name,
 				CCS_MAX_PATHNAME_LEN - 1);
-			ccs_fill_path_info(&r);
+			ccs_fill_path_info(&rn);
 			break;
 		}
 	}
 
 	/* Compare basename of real_program_name and argv[0] */
-	if (bprm->argc > 0 && ccs_check_flags(CCS_TOMOYO_MAC_FOR_ARGV0)) {
+	r->mode = ccs_check_flags(r->domain, CCS_TOMOYO_MAC_FOR_ARGV0);
+	if (bprm->argc > 0 && r->mode) {
 		char *base_argv0 = tmp->buffer;
 		const char *base_filename;
 		retval = -ENOMEM;
@@ -1091,8 +1071,10 @@ static int find_next_domain(struct linux_binprm *bprm,
 		else
 			base_filename++;
 		if (strcmp(base_argv0, base_filename)) {
-			retval = ccs_check_argv0_perm(&r, base_argv0);
-			if (retval)
+			retval = ccs_check_argv0_perm(r, &rn, base_argv0);
+			if (retval == 1)
+				goto retry;
+			if (retval < 0)
 				goto out;
 		}
 	}
@@ -1103,37 +1085,40 @@ static int find_next_domain(struct linux_binprm *bprm,
 		/* Is this program allowed to be aggregated? */
 		list1_for_each_entry(ptr, &aggregator_list, list) {
 			if (ptr->is_deleted ||
-			    !ccs_path_matches_pattern(&r, ptr->original_name))
+			    !ccs_path_matches_pattern(&rn, ptr->original_name))
 				continue;
 			memset(real_program_name, 0, CCS_MAX_PATHNAME_LEN);
 			strncpy(real_program_name, ptr->aggregated_name->name,
 				CCS_MAX_PATHNAME_LEN - 1);
-			ccs_fill_path_info(&r);
+			ccs_fill_path_info(&rn);
 			break;
 		}
 	}
 
 	/* Check execute permission. */
-	retval = ccs_check_exec_perm(&r, bprm, tmp);
+	r->mode = mode;
+	retval = ccs_check_exec_perm(r, &rn);
+	if (retval == 1)
+		goto retry;
 	if (retval < 0)
 		goto out;
 
  calculate_domain:
 	new_domain_name = tmp->buffer;
-	if (is_domain_initializer(old_domain->domainname, &r, &l)) {
+	if (is_domain_initializer(r->domain->domainname, &rn, &ln)) {
 		/* Transit to the child of KERNEL_DOMAIN domain. */
 		snprintf(new_domain_name, CCS_MAX_PATHNAME_LEN + 1,
 			 ROOT_NAME " " "%s", real_program_name);
-	} else if (old_domain == &KERNEL_DOMAIN && !sbin_init_started) {
+	} else if (r->domain == &KERNEL_DOMAIN && !sbin_init_started) {
 		/*
 		 * Needn't to transit from kernel domain before starting
 		 * /sbin/init. But transit from kernel domain if executing
 		 * initializers because they might start before /sbin/init.
 		 */
-		domain = old_domain;
-	} else if (is_domain_keeper(old_domain->domainname, &r, &l)) {
+		domain = r->domain;
+	} else if (is_domain_keeper(r->domain->domainname, &rn, &ln)) {
 		/* Keep current domain. */
-		domain = old_domain;
+		domain = r->domain;
 	} else {
 		/* Normal domain transition. */
 		snprintf(new_domain_name, CCS_MAX_PATHNAME_LEN + 1,
@@ -1144,42 +1129,51 @@ static int find_next_domain(struct linux_binprm *bprm,
 	domain = ccs_find_domain(new_domain_name);
 	if (domain)
 		goto done;
-	if (is_enforce && ccs_check_supervisor("#Need to create domain\n%s\n",
-					       new_domain_name))
+	if (is_enforce) {
+		int error = ccs_check_supervisor(r,
+						 "# wants to create domain\n"
+						 "%s\n", new_domain_name);
+		if (error == 1)
+			goto retry;
+		if (error < 0)
 			goto done;
-	domain = ccs_find_or_assign_new_domain(new_domain_name,
-					       old_domain->profile);
+	}
+	domain = ccs_find_or_assign_new_domain(new_domain_name, r->profile);
 	if (domain)
-		audit_domain_creation_log(new_domain_name, mode,
-					  domain->profile);
+		audit_domain_creation_log(domain);
  done:
 	if (!domain) {
 		printk(KERN_WARNING "TOMOYO-ERROR: Domain '%s' not defined.\n",
 		       new_domain_name);
 		if (is_enforce)
 			retval = -EPERM;
+		else {
+			retval = 0;
+			ccs_set_domain_flag(r->domain, false,
+					    DOMAIN_FLAGS_TRANSITION_FAILED);
+		}
 	} else {
 		retval = 0;
 	}
  out:
 	ccs_free(real_program_name);
 	ccs_free(symlink_program_name);
-	*next_domain = domain ? domain : old_domain;
+	if (domain)
+		r->domain = domain;
 	return retval;
 }
 
 /**
  * check_environ - Check permission for environment variable names.
  *
- * @bprm: Pointer to "struct linux_binprm".
- * @tmp:  Buffer for temporal use.
+ * @r: Pointer to "struct ccs_request_info".
  *
  * Returns 0 on success, negative value otherwise.
  */
-static int check_environ(struct linux_binprm *bprm, struct ccs_page_buffer *tmp)
+static int check_environ(struct ccs_request_info *r)
 {
-	const u8 profile = current->domain_info->profile;
-	const u8 mode = ccs_check_flags(CCS_TOMOYO_MAC_FOR_ENV);
+	struct linux_binprm *bprm = r->bprm;
+	struct ccs_page_buffer *tmp = r->obj->tmp;
 	char *arg_ptr = tmp->buffer;
 	int arg_len = 0;
 	unsigned long pos = bprm->p;
@@ -1189,7 +1183,7 @@ static int check_environ(struct linux_binprm *bprm, struct ccs_page_buffer *tmp)
 	int envp_count = bprm->envc;
 	/* printk(KERN_DEBUG "start %d %d\n", argv_count, envp_count); */
 	int error = -ENOMEM;
-	if (!mode || !envp_count)
+	if (!r->mode || !envp_count)
 		return 0;
 	while (error == -ENOMEM) {
 		struct page *page;
@@ -1239,7 +1233,7 @@ static int check_environ(struct linux_binprm *bprm, struct ccs_page_buffer *tmp)
 			}
 			if (c)
 				continue;
-			if (ccs_check_env_perm(arg_ptr, profile, mode)) {
+			if (ccs_check_env_perm(r, arg_ptr)) {
 				error = -EPERM;
 				break;
 			}
@@ -1259,7 +1253,7 @@ static int check_environ(struct linux_binprm *bprm, struct ccs_page_buffer *tmp)
 		offset = 0;
 	}
  out:
-	if (error && mode != 3)
+	if (r->mode != 3)
 		error = 0;
 	return error;
 }
@@ -1277,7 +1271,10 @@ static void unescape(unsigned char *dest)
 	unsigned char c;
 	unsigned char d;
 	unsigned char e;
-	while ((c = *src++) != '\0') {
+	while (1) {
+		c = *src++;
+		if (!c)
+			break;
 		if (c != '\\') {
 			*dest++ = c;
 			continue;
@@ -1312,10 +1309,7 @@ static inline int root_depth(struct dentry *dentry, struct vfsmount *vfsmnt)
 {
 	int depth = 0;
 	/***** CRITICAL SECTION START *****/
-	spin_lock(&dcache_lock);
-#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 5, 0)
-	spin_lock(&vfsmount_lock);
-#endif
+	ccs_realpath_lock();
 	for (;;) {
 		if (dentry == vfsmnt->mnt_root || IS_ROOT(dentry)) {
 			/* Global root? */
@@ -1328,10 +1322,7 @@ static inline int root_depth(struct dentry *dentry, struct vfsmount *vfsmnt)
 		dentry = dentry->d_parent;
 		depth++;
 	}
-#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 5, 0)
-	spin_unlock(&vfsmount_lock);
-#endif
-	spin_unlock(&dcache_lock);
+	ccs_realpath_unlock();
 	/***** CRITICAL SECTION END *****/
 	return depth;
 }
@@ -1375,18 +1366,14 @@ static int get_root_depth(void)
 /**
  * try_alt_exec - Try to start execute handler.
  *
- * @bprm:        Pointer to "struct linux_binprm".
- * @filename:    The name of requested program.
- * @work:        Pointer to pointer to the name of execute handler.
- * @next_domain: Pointer to pointer to "struct domain_info".
- * @tmp:         Buffer for temporal use.
+ * @r:           Pointer to "struct ccs_request_info".
+ * @handler:     Pointer to the name of execute handler.
+ * @eh_path:     Pointer to pointer to the name of execute handler.
  *
  * Returns 0 on success, negative value otherwise.
  */
-static int try_alt_exec(struct linux_binprm *bprm,
-			const struct path_info *filename, char **work,
-			struct domain_info **next_domain,
-			struct ccs_page_buffer *tmp)
+static int try_alt_exec(struct ccs_request_info *r,
+			const struct path_info *handler, char **eh_path)
 {
 	/*
 	 * Contents of modified bprm.
@@ -1424,18 +1411,19 @@ static int try_alt_exec(struct linux_binprm *bprm,
 	 * modified bprm->argv[bprm->envc + bprm->argc + 6]
 	 *     = original bprm->envp[bprm->envc - 1]
 	 */
+	struct linux_binprm *bprm = r->bprm;
 	struct file *filp;
 	int retval;
 	const int original_argc = bprm->argc;
 	const int original_envc = bprm->envc;
 	struct task_struct *task = current;
-	char *buffer = tmp->buffer;
+	char *buffer = r->obj->tmp->buffer;
 	/* Allocate memory for execute handler's pathname. */
 	char *execute_handler = ccs_alloc(sizeof(struct ccs_page_buffer));
-	*work = execute_handler;
+	*eh_path = execute_handler;
 	if (!execute_handler)
 		return -ENOMEM;
-	strncpy(execute_handler, filename->name,
+	strncpy(execute_handler, handler->name,
 		sizeof(struct ccs_page_buffer) - 1);
 	unescape(execute_handler);
 
@@ -1497,7 +1485,7 @@ static int try_alt_exec(struct linux_binprm *bprm,
 			 "pid=%d uid=%d gid=%d euid=%d egid=%d suid=%d "
 			 "sgid=%d fsuid=%d fsgid=%d state[0]=%u "
 			 "state[1]=%u state[2]=%u",
-			 task->pid, task->uid, task->gid, task->euid,
+			 (pid_t) sys_getpid(), task->uid, task->gid, task->euid,
 			 task->egid, task->suid, task->sgid, task->fsuid,
 			 task->fsgid, (u8) (tomoyo_flags >> 24),
 			 (u8) (tomoyo_flags >> 16), (u8) (tomoyo_flags >> 8));
@@ -1561,7 +1549,7 @@ static int try_alt_exec(struct linux_binprm *bprm,
 	if (retval < 0)
 		goto out;
 	task->tomoyo_flags |= CCS_DONT_SLEEP_ON_ENFORCE_ERROR;
-	retval = find_next_domain(bprm, next_domain, filename, tmp);
+	retval = find_next_domain(r, handler);
 	task->tomoyo_flags &= ~CCS_DONT_SLEEP_ON_ENFORCE_ERROR;
  out:
 	return retval;
@@ -1595,6 +1583,84 @@ static const struct path_info *find_execute_handler(const u8 type)
 	return NULL;
 }
 
+/* List of next_domain which is used for checking interpreter's permissions. */
+struct execve_entry {
+	struct list_head list;
+	struct task_struct *task;
+	struct domain_info *next_domain;
+};
+
+static LIST_HEAD(execve_list);
+static DEFINE_SPINLOCK(execve_list_lock);
+
+/**
+ * ccs_register_next_domain - Remember next_domain.
+ *
+ * @next_domain: Pointer to "struct domain_info".
+ *
+ * Returns 0 on success, -ENOMEM otherwise.
+ */
+static int ccs_register_next_domain(struct domain_info *next_domain)
+{
+	struct execve_entry *ee = kmalloc(sizeof(*ee), GFP_KERNEL);
+	if (!ee)
+		return -ENOMEM;
+	ee->task = current;
+	ee->next_domain = next_domain;
+	/***** CRITICAL SECTION START *****/
+	spin_lock(&execve_list_lock);
+	list_add(&ee->list, &execve_list);
+	spin_unlock(&execve_list_lock);
+	/***** CRITICAL SECTION END *****/
+	return 0;
+}
+
+/**
+ * ccs_fetch_next_domain - Fetch next_domain from the list.
+ *
+ * Returns pointer to "struct domain_info" which will be used if execve()
+ * succeeds. This function does not return NULL.
+ */
+struct domain_info *ccs_fetch_next_domain(void)
+{
+	struct task_struct *task = current;
+	struct domain_info *next_domain = task->domain_info;
+	struct execve_entry *p;
+	/***** CRITICAL SECTION START *****/
+	spin_lock(&execve_list_lock);
+	list_for_each_entry(p, &execve_list, list) {
+		if (p->task != task)
+			continue;
+		next_domain = p->next_domain;
+		break;
+	}
+	spin_unlock(&execve_list_lock);
+	/***** CRITICAL SECTION END *****/
+	return next_domain;
+}
+
+/**
+ * ccs_unregister_next_domain - Forget next_domain.
+ */
+static void ccs_unregister_next_domain(void)
+{
+	struct task_struct *task = current;
+	struct execve_entry *p;
+	struct execve_entry *ee = NULL;
+	/***** CRITICAL SECTION START *****/
+	spin_lock(&execve_list_lock);
+	list_for_each_entry(p, &execve_list, list) {
+		if (p->task != task)
+			continue;
+		list_del(&p->list);
+		ee = p;
+		break;
+	}
+	spin_unlock(&execve_list_lock);
+	/***** CRITICAL SECTION END *****/
+	kfree(ee);
+}
+
 /**
  * search_binary_handler_with_transition - Perform domain transition.
  *
@@ -1607,59 +1673,76 @@ static const struct path_info *find_execute_handler(const u8 type)
 int search_binary_handler_with_transition(struct linux_binprm *bprm,
 					  struct pt_regs *regs)
 {
-	struct task_struct *task = current;
-	struct domain_info *next_domain = NULL;
-	struct domain_info *prev_domain = task->domain_info;
-	const struct path_info *handler;
 	int retval;
+	struct task_struct *task = current;
+	const struct path_info *handler;
+	struct ccs_request_info r;
+	struct obj_info obj;
 	/*
-	 * "work" holds path to program.
+	 * "eh_path" holds path to execute handler program.
 	 * Thus, keep valid until search_binary_handler() finishes.
 	 */
-	char *work = NULL;
-	struct ccs_page_buffer *buf = ccs_alloc(sizeof(struct ccs_page_buffer));
-	ccs_load_policy(bprm->filename);
-	if (!buf)
+	char *eh_path = NULL;
+	struct ccs_page_buffer *tmp = ccs_alloc(sizeof(struct ccs_page_buffer));
+	memset(&obj, 0, sizeof(obj));
+	if (!sbin_init_started)
+		ccs_load_policy(bprm->filename);
+	if (!tmp)
 		return -ENOMEM;
-	/* printk(KERN_DEBUG "rootdepth=%d\n", get_root_depth()); */
+
+	ccs_init_request_info(&r, NULL, CCS_TOMOYO_MAC_FOR_FILE);
+	r.bprm = bprm;
+	r.obj = &obj;
+	obj.path1_dentry = bprm->file->f_dentry;
+	obj.path1_vfsmnt = bprm->file->f_vfsmnt;
+	obj.tmp = tmp;
+
+	/* Clear manager flag. */
+	task->tomoyo_flags &= ~CCS_TASK_IS_POLICY_MANAGER;
 	handler = find_execute_handler(TYPE_EXECUTE_HANDLER);
 	if (handler) {
-		retval = try_alt_exec(bprm, handler, &work, &next_domain, buf);
+		retval = try_alt_exec(&r, handler, &eh_path);
 		if (!retval)
-			audit_execute_handler_log(true, work, bprm);
+			audit_execute_handler_log(true, handler->name, bprm);
 		goto ok;
 	}
-	retval = find_next_domain(bprm, &next_domain, NULL, buf);
+	retval = find_next_domain(&r, NULL);
 	if (retval != -EPERM)
 		goto ok;
 	handler = find_execute_handler(TYPE_DENIED_EXECUTE_HANDLER);
 	if (handler) {
-		retval = try_alt_exec(bprm, handler, &work, &next_domain, buf);
+		retval = try_alt_exec(&r, handler, &eh_path);
 		if (!retval)
-			audit_execute_handler_log(false, work, bprm);
+			audit_execute_handler_log(false, handler->name, bprm);
 	}
  ok:
-	if (retval)
+	if (retval < 0)
 		goto out;
-	task->domain_info = next_domain;
-	retval = check_environ(bprm, buf);
-	if (retval)
+	r.mode = ccs_check_flags(r.domain, CCS_TOMOYO_MAC_FOR_ENV);
+	retval = check_environ(&r);
+	if (retval < 0)
+		goto out;
+	retval = ccs_register_next_domain(r.domain);
+	if (retval < 0)
 		goto out;
 	task->tomoyo_flags |= TOMOYO_CHECK_READ_FOR_OPEN_EXEC;
 	retval = search_binary_handler(bprm, regs);
 	task->tomoyo_flags &= ~TOMOYO_CHECK_READ_FOR_OPEN_EXEC;
- out:
-	/* Return to previous domain if execution failed. */
 	if (retval < 0)
-		task->domain_info = prev_domain;
+		goto out;
+	/* Proceed to next domain if execution suceeded. */
+	task->domain_info = r.domain;
+	mb(); /* Make domain transition visible to other CPUs. */
 	/* Mark the current process as execute handler. */
-	else if (handler)
+	if (handler)
 		task->tomoyo_flags |= TOMOYO_TASK_IS_EXECUTE_HANDLER;
 	/* Mark the current process as normal process. */
 	else
 		task->tomoyo_flags &= ~TOMOYO_TASK_IS_EXECUTE_HANDLER;
-	ccs_free(work);
-	ccs_free(buf);
+ out:
+	ccs_unregister_next_domain();
+	ccs_free(eh_path);
+	ccs_free(tmp);
 	return retval;
 }
 
@@ -1677,6 +1760,8 @@ int search_binary_handler_with_transition(struct linux_binprm *bprm,
 					  struct pt_regs *regs)
 {
 #ifdef CONFIG_SAKURA
+	/* Clear manager flag. */
+	current->tomoyo_flags &= ~CCS_TASK_IS_POLICY_MANAGER;
 	ccs_load_policy(bprm->filename);
 #endif
 	return search_binary_handler(bprm, regs);
